{
    This file is part of FloatView.
    Copyright (c) 2003-2008, Tom Verhoeff (TU/e).

    FloatView provides access to the bit-level format of the
    Single and Double floating-point types, based on
    ANSI/IEEE Std 754-1985 for Binary Floating-Point Arithmetic.

    See the file COPYING.FloatView, included in this distribution,
    for details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************}

unit FloatView;

interface

const
  SignIndexSingle =    31; { position of sign bit in single format }
  ExpSizeSingle   =     8; { size of exponent field in single format }
  MaxExpSingle    =   1 shl ExpSizeSingle - 1; { largest biased exponent }
  ExpBiasSingle   =   127; { bias of exponent in single format }
  EminSingle      = 1 - ExpBiasSingle; { smallest exponent in single format }
  EmaxSingle      = MaxExpSingle - 1 - ExpBiasSingle; { largest exponent }
  FracSizeSingle  =    23; { size of fraction field in single format }

type
  BitIndexSingle = 0 .. SignIndexSingle;
  ConvertSingle = record
    case Boolean of
      False: ( fp: Single ) { IEEE 754 floating-point number in single format }
    ; True : ( s: Cardinal ) { sequence of bits }
      { bit i of fp = ord ( ((Cardinal(1) shl i) and s) <> 0 ) }
    end;

function SingleToStrBits ( const x: Single ): String;
  { returns string with bits for sign, biased exponent, and fraction }
function SingleToStrBinary ( const x: Single ): String;
  { returns string with binary scientific representation of x:
    NaN, +/-Infinity, +/-B.B___Bd+/-E, where
    'B' is a bit, 'd' stands for '*2^' and 'E' is the exponent in decimal }
function PowerOf2Single ( const x: Single; n: Integer ): Single;
  { returns x * 2 ^ n }
procedure FlipBitSingle ( var x: Single; i: BitIndexSingle );
  { flips bit i of x }
function SuccSingle ( const x: Single): Single;
  { returns successor of x }
function PredSingle ( const x: Single): Single;
  { returns predecessor of x }

const
  SignIndexDouble =    63; { position of sign bit in double format }
  ExpSizeDouble   =    11; { size of exponent field in double format }
  MaxExpDouble    =  1 shl ExpSizeDouble - 1; { largest biased exponent }
  ExpBiasDouble   =  1023; { bias of exponent in double format }
  EminDouble      = 1 - ExpBiasDouble; { smallest exponent in double format }
  EmaxDouble      = MaxExpDouble - 1 - ExpBiasDouble; { largest exponent }
  FracSizeDouble  =    52; { size of fraction field in double format }

type
  BitIndexDouble = 0 .. SignIndexDouble;
  ConvertDouble = record
    case Boolean of
      False: ( fp: Double ) { IEEE 754 floating-point number in single format }
    ; True : ( s: QWord ) { sequence of bits }
      { bit i of fp = ord ( ((QWord(1) shl i) and s) <> 0 ) }
    end;

function DoubleToStrBits ( const x: Double ): String;
  { returns string with bits for sign, biased exponent, and fraction }
function DoubleToStrBinary ( const x: Double ): String;
  { returns string with binary scientific representation of x:
    NaN, +/-Infinity, +/-B.B___Bd+/-E, where
    'B' is a bit, 'd' stands for '*2^' and 'E' is the exponent in decimal }
function PowerOf2Double ( const x: Double; n: Integer ): Double;
  { returns x * 2 ^ n }
procedure FlipBitDouble ( var x: Double; i: BitIndexDouble );
  { flips bit i of x }
function SuccDouble ( const x: Double): Double;
  { returns successor of x }
function PredDouble ( const x: Double): Double;
  { returns predecessor of x }

implementation

const
  FracExpSep = 'd'; { separates fraction from exponent in strings }

function DigitToStr ( n: Integer ): String;
  { returns string representation of digit n }
  { pre: 0 <= n <= 9 }
  begin
    DigitToStr := chr ( ord ( '0' ) + n )
  end; { DigitToStr }

function IntToStr ( n: Integer; w: Integer ): String;
  { returns signed decimal representation of i in w digits, left-padded with zero }
  { pre: 1 <= w }
  const
    Base = 10; { base of conversion }
  var
    sign: String; { sign }
    r: String; { unsigned result }
  begin
    if n < 0 then sign := '-'
    else sign := '+'
  ; n := abs ( n )
  ; r := ''

  ; while w <> 0 do
    begin
      r := DigitToStr ( n mod Base ) + r
    ; n := n div Base
    ; w := w - 1
    end { while }

  ; IntToStr := sign + r
  end; { IntToStr }

function SingleToStrBits ( const x: Single ): String;
  { returns string with bits for sign, biased exponent, and fraction }
  var
    c: ConvertSingle; { makes bits of x accessible }
    i: BitIndexSingle; { traverses bits of x }
    r: String; { result }
  begin
    r := ''
  ; with c do
    begin
      fp := x

    ; for i := SignIndexSingle downto 0 do
      begin
        if ((Cardinal(1) shl i) and s) <> 0 then r := r + '1'
        else r := r + '0'
      ; if i in [ SignIndexSingle, SignIndexSingle - ExpSizeSingle ] then
          r := r + '_'
      end { for i }

    end { with c }
  ; SingleToStrBits := r
  end; { SingleToStrBits }

function ExpFieldSingle ( const x: Single ): Integer;
  { returns biased exponent of x }
  var
    c: ConvertSingle; { makes bits of x accessible }
    i: BitIndexSingle; { traverses bits of x }
    e: Integer; { result }
  begin
    with c do
    begin
      fp := x
    ; e := 0
  
    ; for i := SignIndexSingle - 1 downto
               SignIndexSingle - ExpSizeSingle do
      begin
        e := 2 * e + ord ( ((Cardinal(1) shl i) and s) <> 0 )
      end { for i }

    ; ExpFieldSingle := e
    end { with c }
  end; { ExpFieldSingle }

function FracFieldZeroSingle ( const x: Single ): Boolean;
  { returns whether fraction field of x is all zero }
  var
    c: ConvertSingle; { makes bits of x accessible }
    i, j: BitIndexSingle; { traverses bits of x }
  begin
    with c do
    begin
      fp := x
    ; i := 0
    ; j := SignIndexSingle - ExpSizeSingle

    ; { Bounded Linear Search for nonzero }
      while i <> j do
      begin
        if ((Cardinal(1) shl i) and s) <> 0 then j := i
        else i := i + 1
      end { while }

    ; FracFieldZeroSingle := i = SignIndexSingle - ExpSizeSingle
    end { with c }
  end; { FracFieldZeroSingle }

function SingleToStrBinary ( const x: Single ): String;
  { returns string with binary scientific representation of x:
    NaN, +/-Infinity, +/-B.B___Bd+/-E, where
    'B' is a bit, 'd' stands for '*2^' and 'E' is the exponent in decimal }
  const
    Width = 3 + FracSizeSingle + 5;
      { for sign, bit, '.', fraction bits, separator, signed exponent }
  var
    c: ConvertSingle; { makes bits of x accessible }
    i: BitIndexSingle; { traverses bits x }
    e: Integer; { exponent field }
    z: Boolean; { whether fraction field is all zero }
    r: String; { result }
  begin
    r := ''
  ; with c do
    begin
      fp := x
    ; e := ExpFieldSingle ( x )
    ; z := FracFieldZeroSingle ( x )

    ; if ( e = MaxExpSingle ) and not z then
      begin { NaN }
        r := r + 'NaN'
      end { then }
      else { not NaN }
      begin
        { convert sign bit }
        if ((Cardinal(1) shl SignIndexSingle) and s) <> 0 then r := r + '-'
        else r := r + '+'
        { case analysis on biased exponent }
      ; if e = MaxExpSingle then { z }
        begin { infinity }
          r := r + 'Infinity'
        end { then }
        else if ( e = 0 ) and z then
        begin { zero }
          r := r + '0.0'
        end { then }
        else
        begin
          if e = 0 then { not z, subnormal (denormalized) }
            r := r + '0.'
          else { 0 < e < MaxExpSingle, normalized }
            r := r + '1.'

          { convert fraction }
        ; for i := SignIndexSingle - ExpSizeSingle - 1 downto 0 do
          begin
            r := r + DigitToStr ( ord ( ((Cardinal(1) shl i) and s) <> 0 ) )
          end { for i }

          { convert biased exponent }
        ; if e = 0 then { subnormal }
              r := r + FracExpSep + IntToStr ( EminSingle, 3 )
//        else if e - ExpBiasSingle = 0 then { no exponent if it is zero }
          else r := r + FracExpSep + IntToStr ( e - ExpBiasSingle, 3 )
        end { else }
      end { else }

    end { with c }
  ; SingleToStrBinary := r
  end; { SingleToStrBinary }

function PowerOf2Single ( const x: Single; n: Integer ): Single;
  { returns x * 2 ^ n }
  var
    r: Single; { result }
    k: Integer; { exponent }
  begin
    r := x
  ; k := 0 {@ inv: r = x * 2 ^ k }

  ; while k <> n do
    begin { n <> 0 }
      if n > 0 then
      begin
        r := 2 * r
      ; k := k + 1
      end { then }
      else { n < 0 }
      begin
        r := r / 2
      ; k := k - 1
      end { else }
    end { while }

  ; PowerOf2Single := r
  end; { PowerOf2Single }

procedure FlipBitSingle ( var x: Single; i: BitIndexSingle );
  { flips bit i of x }
  var
    c: ConvertSingle; { makes bits of x accessible }
  begin
    with c do
    begin
      fp := x
    ; s := s xor (QWord(1) shl i)
    ; x := fp
    end { with c }
  end; { FlipBitSingle }

function SuccSingle ( const x: Single): Single;
  { returns successor of x }
  var
    c: ConvertSingle; { makes bits of x accessible }
  begin
    with c do
    begin
      fp := x
    ; Inc ( s )
    ; SuccSingle := fp
    end { with c }
  end; { SuccSingle }

function PredSingle ( const x: Single): Single;
  { returns predecessor of x }
  var
    c: ConvertSingle; { makes bits of x accessible }
  begin
    with c do
    begin
      fp := x
    ; Dec ( s )
    ; PredSingle := fp
    end { with c }
  end; { PredSingle }

function DoubleToStrBits ( const x: Double ): String;
  { returns string with bits for sign, biased exponent, and fraction }
  var
    c: ConvertDouble; { makes bits of x accessible }
    i: BitIndexDouble; { traverses bits of x }
    r: String; { result }
  begin
    r := ''
  ; with c do
    begin
      fp := x

    ; for i := SignIndexDouble downto 0 do
      begin
      ; if ((QWord(1) shl i) and s) <> 0 then r := r + '1'
        else r := r + '0'
      ; if i in [ SignIndexDouble, SignIndexDouble - ExpSizeDouble ] then
          r := r + '_'
      end { for i }

    end { with c }
  ; DoubleToStrBits := r
  end; { DoubleToStrBits }

function ExpFieldDouble ( const x: Double ): Integer;
  { returns biased exponent of x }
  var
    c: ConvertDouble; { makes bits of x accessible }
    i: BitIndexDouble; { traverses bits of x }
    e: Integer; { result }
  begin
    with c do
    begin
      fp := x
    ; e := 0
  
    ; for i := SignIndexDouble - 1 downto
               SignIndexDouble - ExpSizeDouble do
      begin
      ; e := 2 * e + ord ( ((QWord(1) shl i) and s) <> 0 )
      end { for i }

    ; ExpFieldDouble := e
    end { with c }
  end; { ExpFieldDouble }

function FracFieldZeroDouble ( const x: Double ): Boolean;
  { returns whether fraction field of x is all zero }
  var
    c: ConvertDouble; { makes bits of x accessible }
    i, j: BitIndexDouble; { traverses bits of x }
  begin
    with c do
    begin
      fp := x
    ; i := 0
    ; j := SignIndexDouble - ExpSizeDouble

    ; { Bounded Linear Search for nonzero }
      while i <> j do
      begin
      ; if ((QWord(1) shl i) and s) <> 0 then j := i
        else i := i + 1
      end { while }

    ; FracFieldZeroDouble := i = SignIndexDouble - ExpSizeDouble
    end { with c }
  end; { FracFieldZeroDouble }

function DoubleToStrBinary ( const x: Double ): String;
  { returns string with binary scientific representation of x:
    NaN, +/-Infinity, +/-B.B___Bd+/-E, where
    'B' is a bit, 'd' stands for '*2^' and 'E' is the exponent in decimal }
  const
    Width = 3 + FracSizeDouble + 6; { sign, bit, '.', fraction bits, separator, signed exponent }
  var
    c: ConvertDouble; { makes bits of x accessible }
    i: BitIndexDouble; { traverses bits of x }
    e: Integer; { exponent field }
    z: Boolean; { whether fraction field is all zero }
    r: String; { result }
  begin
    r := ''
  ; with c do
    begin
      fp := x
    ; e := ExpFieldDouble ( x )
    ; z := FracFieldZeroDouble ( x )

    ; if ( e = MaxExpDouble ) and not z then
      begin { NaN }
        r := r + 'NaN'
      end { then }
      else { not NaN }
      begin
        { convert sign bit }
        if ((QWord(1) shl SignIndexDouble) and s) <> 0 then r := r + '-'
        else r := r + '+'
        { case analysis on biased exponent e }
      ; if e = MaxExpDouble then
        begin { infinity }
          r := r + 'Infinity'
        end { then }
        else if ( e = 0 ) and z then
        begin { zero }
          r := r + '0.0'
        end { then }
        else
        begin
          if e = 0 then { subnormal (denormalized) }
            r := r + '0.'
          else { normalized }
            r := r +'1.'

          { convert fraction }
        ; for i := SignIndexDouble - ExpSizeDouble - 1 downto 0 do
          begin
          ; r := r + DigitToStr ( ord ( ((QWord(1) shl i) and s) <> 0 ) )
          end { for i }

          { convert biased exponent }
        ; if e = 0 then r := r + FracExpSep + IntToStr ( EminDouble, 4 )
//          else if e - ExpBiasDouble = 0 then { skip }
          else r := r + FracExpSep + IntToStr ( e - ExpBiasDouble, 4 )
        end { else }
      end { else }

    end { with c }
  ; DoubleToStrBinary := r
  end; { DoubleToStrBinary }

function PowerOf2Double ( const x: Double; n: Integer ): Double;
  { returns x * 2 ^ n }
  var
    r: Double; { result }
    k: Integer; { exponent }
  begin
    r := x
  ; k := 0 {@ inv: r = x * 2 ^ k }

  ; while k <> n do
    begin { n <> 0 }
      if n > 0 then
      begin
        r := 2 * r
      ; k := k + 1
      end { then }
      else { n < 0 }
      begin
        r := r / 2
      ; k := k - 1
      end { else }
    end { while }

  ; PowerOf2Double := r
  end; { PowerOf2Double }

procedure FlipBitDouble ( var x: Double; i: BitIndexDouble );
  { flips bit i of x }
  var
    c: ConvertDouble; { makes bits of x accessible }
  begin
    with c do
    begin
      fp := x
    ; s := s xor (QWord(1) shl i)
    ; x := fp
    end { with c }
  end; { FlipBitDouble }

function SuccDouble ( const x: Double): Double;
  { returns successor of x }
  var
    c: ConvertDouble; { makes bits of x accessible }
  begin
    with c do
    begin
      fp := x
    ; Inc ( s )
    ; SuccDouble := fp
    end { with c }
  end; { SuccDouble }

function PredDouble ( const x: Double): Double;
  { returns predecessor of x }
  var
    c: ConvertDouble; { makes bits of x accessible }
  begin
    with c do
    begin
      fp := x
    ; Dec ( s )
    ; PredDouble := fp
    end { with c }
  end; { PredDouble }

end.
